import ast
import base64
import datetime
import decimal
import json
import logging
import os
import re
import sqlparse
from urllib.parse import urlparse

import boto3
from botocore.exceptions import ClientError
from os.path import dirname, basename
from boto3.dynamodb.conditions import Key, Attr

logger = logging.getLogger()
logger.setLevel(os.getenv("log_level") or logging.INFO)

# Prerequisite is DynamoDB table 'glueHistory' must be created with partition key 'database' which is unique to all database/tables and sort key 'table'

# Create table workflow as follows
# 1. User creates table using 'create table as', Glue event triggered
# 2. Event parsed, fetch new Glue record from catalog including S3 location
# 3. Write record into DDB

# Delete table workflow as follows
# 1. User drops table (Glue console, API/CLI, Athena) and specifies key word "--drop-store"
# 2. Athena event parsed, fetch s3_path from DDB using database and table with purged != True (otherwise old table).
# 3. Run async delete from S3 path including manifest files at same level.
# 4. Add deleted attribute to item in DDB

# Notify via SNS on either create or delete

# ---------------------
# ENVIRONMENT VARIABLES
# ---------------------
STORE_DROP_KEYWORD = os.environ['Athena_SQL_Drop_Phrase']
DDB_HISTORY_TABLE = os.environ['ddb_history_table']
DISABLE_S3_CLEANUP = os.environ["disable_s3_cleanup"]
DB_LIST=os.environ["db_list"]

# END OF ENVIRONMENT VARIABLES


# Helper class to convert a DynamoDB item to JSON.
class DecimalEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            if o % 1 > 0:
                return float(o)
            else:
                return int(o)
        return super(DecimalEncoder, self).default(o)

def get_glue_record(database, table):
    """
    Description:
    --
    Get the metadata about the glue table and if a view definition exists, get
    the view
    :param database: name of the glue database in which table exists
    :param table: name fo the glue table
    :return: List of table metadata
    """
    gluecat = boto3.client('glue')
    response = gluecat.get_table(
        DatabaseName=database,
        Name=table
        )
    if response:
        table_md = response['Table']
        view_sql = table_md.get('ViewOriginalText')
        if view_sql:  # sometimes must decode base64 text
            try:
                view_sql = view_sql.replace('/* Presto View: ', '') \
                    .replace(' */', '')
                view_sql = base64.b64decode(view_sql).decode('utf-8')
                view_sql = json.loads(view_sql)['originalSql']
            except Exception:
                pass  # accept view_sql as is
#
        values = [
            table_md['Name'],
            table_md['DatabaseName'],
            table_md.get('Owner'),
            table_md.get('CreatedBy'),
            table_md.get('CreateTime').strftime('%Y-%m-%d %H:%M:%S'),
            table_md.get('StorageDescriptor').get('Location') or None,
            view_sql
        ]
        return values
    else:
        return none

def get_ddb_history_record(database,table):
    """
    Description:
    --
    Return the s3_path and table_date for given table and database from
    DynamoDb History Table (DDB_HISTORY_TABLE)
    Query returns last entry sorted by create_time
    :param database: Glue Database Name
    :param table: Glue table that is being dropped
    :return: s3_path, S3 data location for the external table being dropped
    """
    dynamodb = boto3.resource('dynamodb')
    ddb_table = dynamodb.Table(DDB_HISTORY_TABLE)
    s3_path = None
    logger.debug("table is %s.%s", database, table)
    kc = KeyConditionExpression=Key('database').eq(database) & Key('table_date').begins_with(table)

    try:
        response = ddb_table.query(
            ConsistentRead=True,
            KeyConditionExpression=kc,
            Limit=1,
            ScanIndexForward=False
        )
    except ClientError as e:
        print(e.response['Error']['Message'])
    else:
        logger.debug("RESPONSE IS ------ %s", response)
        item = response['Items']
        logger.debug("glueHistory item %s", item)
        s3_path = item[0]['s3_path']
        table_date = item[0]['table_date']
        print(json.dumps(item, indent=4, cls=DecimalEncoder))
        return(s3_path, table_date)

def record_ddb_table_activity(table, database, owner, created_by, create_time, s3_path, view_sql):
    """
    Description:
    --
    Insert a record into the history table for an external table being created in Athena
    :param database: name of the database in which table is being created
    :param table_date: unique name of the table being created by appending timestamp
    :param table: table name
    :param owner: account creating the table
    :param created_by:  user or the ec2 instance from which the table is being created
    :param create_time: table creation time in UTC
    :param s3_path: The s3 data location on top which the table is being defined
    :param view_sql: View DDL if it was a view that is being created
    :return: None
    """
    dynamodb = boto3.resource('dynamodb')
    ddb_table = dynamodb.Table(DDB_HISTORY_TABLE)

    ddb_table.put_item(
        Item={
            'database': database,
            'table_date': table + "_" + create_time,
            'table': table,
            'owner': owner,
            'created_by': created_by,
            's3_path': s3_path,
            'view_sql': view_sql,
            'purged' : False,
            'update_time' : create_time
            }
    )

def update_ddb_history_record(database, table, column, new_value):
    """
    Description:
    --
    Update the latest record in the DynamoDB history table (is duplicate database/table)
    :param database: portion of unique identifier of the table
    :param table: name of the table being updated
    :param column: name of the column being updated
    :param new_value: name of the new value for the attribute (okay to update the s3_path itself)
    :param DDB_HISTORY_TABLE: name of the history table in DynamoDB
    :return: None
    """
    dynamodb = boto3.resource('dynamodb')
    ddb_table = dynamodb.Table(DDB_HISTORY_TABLE)

    # First get table_date sort key needed to update the entry
    result=get_ddb_history_record(database,table)
    table_date=result[1]
    response = ddb_table.update_item(
        Key={
            'database': database,
            'table_date': table_date
        },
        ExpressionAttributeNames={
            '#C': column,
            '#UT': 'update_time'
            },
        ExpressionAttributeValues={
            ':c'    : new_value,
            ':u'    : datetime.datetime.now().isoformat()
        },
        UpdateExpression='SET #C = :c, #UT = :u',
        ReturnValues="UPDATED_NEW"
    )

    logger.debug("UpdateItem succeeded:")
    logger.debug(json.dumps(response, indent=4, cls=DecimalEncoder))

def delete_s3_objects(s3_path):
    """
    Description:
    --
    Delete the objects under the specified s3 location
    :param s3_path: s3 location being purged
    :return: True if object deleted successfully or False
    """

    logger.debug("DISABLE_S3_CLEANUP %s:", DISABLE_S3_CLEANUP)

    if DISABLE_S3_CLEANUP != "true":
        o = urlparse(s3_path)
        bucket_name = o.netloc
        logger.debug("bucket name: %s", bucket_name)
        # EX s3_path:  s3://aws-athena-query-results-01234567890-us-west-2/tables/12345678-b910-d131-e141-1516171819/
        # need to parse out just tables/12345678-b910-d131-e141-1516171819, important to eliminate trailing '/' so table files and associated manifest are also deleted
        prefix13 = '/'.join(s3_path.split('/')[3:]).rstrip('/')
        s3 = boto3.resource('s3')
        bucket = s3.Bucket(bucket_name)

        logger.debug("in delete, prefix13 is: %s: ", prefix13)
        # for loop just to log objects
        for obj in bucket.objects.filter(Prefix=prefix13):
            logger.info("S3 object to delete: %s", obj.key)

        res = bucket.objects.filter(Prefix=prefix13).delete()
        if len(res) > 0:
            result = res[0]
            if "Errors" in result:
                logger.error("S3 Object Cleanup failed: {}".format(result['Errors']))
                return False
            else:
                logger.info("S3 Object Cleanup succeeded")
                logger.debug(result["Deleted"])
                return True
        else:
            logger.debug("Nothing to cleanup in S3 bucket")
            return True
    else:
        logger.info("disable_s3_cleanup in Lambda Environment Variable is "
                    "set to True and hence skipping S3 delete")
        return False

def preprocess_query_string(query_string):
    """
    Description:
    --
    Preprocess the query string.
    1. Convert the query string to lower case
    2. Remove extra white space
    3. Move all the comments to the end of string
    :param query_string: query string from user
    :return: processed query string

    Example
    --
    From: "--comment1 \n--comment2 \n --comment3\nDROP TABLE --drop-store \n`mydatabase1`.`mytable1` "
    To: 'drop table mydatabase1.mytable1  --comment1 --comment2 --comment3 --drop-store'
    """
    query_string = re.sub(r'`', '', query_string).lower()
    query_string = re.sub(' +', ' ', query_string)
    parsed=sqlparse.parse(query_string)[0]
    table_name = None
    for token in parsed.tokens:
        if isinstance(token, sqlparse.sql.Identifier):
            table_name = token
    return query_string, str(table_name)

def db_filter_check(database):
    """
    Description:
    ---
    Checks if database regex list specified in Lambda env variable matches databases from events
     :param database: name of database having Glue or Athena event
     :return True if match, False if not
    """

    logger.info("Checking database against filter")
    isMatch=False
    match=''

    for dbFilterExpr in DB_LIST.split(','):
        logger.debug("Database Filter Expression: <%s>", dbFilterExpr)
        try:
            match = re.fullmatch(dbFilterExpr,database)
            if match:
                logger.info("Match found for database <%s> using expression <%s>", database, dbFilterExpr)
                isMatch=True
                break
            else:
                logger.info("No match found for database <%s> using expression <%s>", database, dbFilterExpr)
                isMatch=False
        except re.error:
            logger.debug("Invalid regex: %s", dbFilterExpr)
    return(isMatch)

# Lambda Handler needs to check for Glue and Athena events
def lambda_handler(event, context):
    """
    Description:
    --
    Process the events from Glue for CreateTable and UpdateTable, from Athena containing SQL for drop store keyphrase.
    :param event: used to trigger lambda from Eventbridge
    :param context: info about the Lambda invocation and triggering itself
    :return: none
    """

    logger.info("Logging Event: %s", event)
    eventType = event["source"]
    if eventType == "aws.athena":
        logger.info("athenaEvent found")
        detailType=event["detail-type"]
        logger.debug("detail type is: %s", detailType)
        if event['detail']['requestParameters']:
            # Sometimes queryString is not available due to size restriction in the event
            # Seen this in big view definitions. Need this check to stop the failure
            if "queryString" in event['detail']['requestParameters'].keys():
                query_string = event["detail"]["requestParameters"]["queryString"]
            else:
                query_string = None
            # This check is required when the query comes from an EC2 instance and
            # the user does not specify QueryExecutionContext in boto3 call
            query_execution_context_db = None
            database = None
            if "queryExecutionContext" in event["detail"]["requestParameters"]:
                query_execution_context_db = event["detail"]["requestParameters"]["queryExecutionContext"]["database"]

            logger.debug("queryString is: %s", query_string)

            if query_string:
                query_string, table_string = preprocess_query_string(query_string)
                logger.debug("processed queryString: {}".format(query_string))
                if "drop table" in query_string:
                    table_array = query_string.split()
                    table_string = table_array[2]
                    if "." in table_string:
                        # account for syntax when database is specified
                        # (e.g. "sampledb.mytable1")
                        db_table = table_string.split(".")
                        database = db_table[0]
                        table = db_table[1]
                    else:
                        table = table_string
                        # if the user does not prefix the table name with database name
                        # check if there is a database set in query execution context
                        if query_execution_context_db:
                            database = query_execution_context_db
                        logger.debug("database.table is: %s.%s ", database, table)

                    # Check if database is in list of approved databases to purge
                    if db_filter_check(database):
                        if STORE_DROP_KEYWORD in query_string:
                            result = get_ddb_history_record(database,table)
                            logger.debug("result is:%s", result)
                            s3_path = result[0]
                            if delete_s3_objects(s3_path):
                                update_ddb_history_record(database,table,'purged',True)
                        else:
                            logger.info("No %s specified, skipping S3 delete", STORE_DROP_KEYWORD)
                    else:
                        logger.info("Skipping s3 delete since <{}> is not in the cleanup list. ".format(database))
                else:
                    logger.info("Nothing to do as query is  not a drop request.")
            else:
                logger.info("Nothing to do as queryString is not available in the event.")
        else:
            logger.info("RequestParameters is empty. May be an invalid request or User does not have permission to execute query.")


# Use Glue events for CreateTable and UpdateTable only
    elif eventType == "aws.glue":
        logger.debug("glueEvent found")
        detailType = event["detail-type"]
        logger.debug("detailType is %s", detailType)
        if (detailType == "Glue Data Catalog Database State Change" ) or (detailType == "Glue Data Catalog Table State Change"):
            database_name = event["detail"]["databaseName"]
            type_of_change = event["detail"]["typeOfChange"]
            logger.info("typeOfChange == {}".format(type_of_change))
        # typeOfChange potential values
        #   :CreateTable =  Athena, Spectrum, or Glue console/API
        #   :BatchDeleteTable = Glue Console
        #   :DeleteTable =  Athena drop command or Glue cli
        #   :DropTable =  haven't seen yet
        #   :UpdateTable (only in Table State Change) =  modify some attribute of the table including S3 location
            if type_of_change == "CreateTable":
                table = event["detail"]["changedTables"][0]
                logger.debug("table to create: %s.%s", database_name,table)
                # table_values [ Name, DatabaseName, Owner, CreatedBy, CreateTime, Location, view_sql ]
                table_values = get_glue_record(database_name, table)
                print("table values: ", table_values)
                logger.debug(table_values)
                record_ddb_table_activity(
                    table=table_values[0],
                    database=table_values[1],
                    owner=None if isinstance(table_values[2], str) and table_values[2] == "" else table_values[2],
                    created_by=None if isinstance(table_values[3], str) and table_values[3] == "" else table_values[3],
                    create_time=table_values[4],
                    s3_path=None if isinstance(table_values[5], str) and table_values[5] == "" else table_values[5],
                    view_sql=None if isinstance(table_values[6], str) and table_values[6] == "" else table_values[6]
                )
            elif type_of_change == "UpdateTable":
                table = event["detail"]["tableName"]
                logger.info("table to update: %s.%s", database_name, table)
                table_values = get_glue_record(database_name,table)
                new_s3_path = None if isinstance(table_values[5], str) and table_values[5] == "" else table_values[5]
                update_ddb_history_record(database_name,table,'s3_path',new_s3_path)
    else:
        logger.debug("No EventType")
